// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"os"
	"testing"

	_ "github.com/mattn/go-sqlite3"
	"github.com/orus-io/yago"
	_ "github.com/slicebit/qb/dialects/postgres"

	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
)

const testMaxDBConnections = 10

func GetTestDB(t *testing.T) (*model.Model, *yago.DB) {
	dsn, ok := os.LookupEnv("XBUS_TEST_SQLITE_DSN")
	if ok {
	} else {
		dsn = ":memory:"
	}
	m, db, err := model.NewDB("sqlite3", dsn, testMaxDBConnections, testutils.GetTestLogger(t), 0)
	if err != nil {
		t.Fatalf("Cannot initialize test database: %s", err)
		return &model.Model{}, nil
	}
	if err := db.Metadata.GetQbMetadata().CreateAll(db.Engine); err != nil {
		t.Fatal(err)
	}
	return m, db
}
