// This code was autogenerated from xbus/control.proto, do not edit.
package control

import (
	"context"
	"log"
	"time"

	"github.com/golang/protobuf/proto"
	nats "github.com/nats-io/go-nats"
	bitbucket_org_orus_io_go_xbus_api "bitbucket.org/orus-io/go-xbus/api"
	github_com_nats_rpc_gogo_nrpc "github.com/nats-rpc/gogo-nrpc"
	"github.com/prometheus/client_golang/prometheus"
	"github.com/nats-rpc/gogo-nrpc"
)

// AccountServer is the interface that providers of the service
// Account should implement.
type AccountServer interface {
	List(ctx context.Context, req AccountListRequest) (resp AccountListReply, err error)
	Update(ctx context.Context, req AccountUpdateRequest) (resp bitbucket_org_orus_io_go_xbus_api.Account, err error)
	Delete(ctx context.Context, req AccountIDRequest) (err error)
	RenewAPIKey(ctx context.Context, req AccountIDRequest) (resp bitbucket_org_orus_io_go_xbus_api.Account, err error)
	GetAPIKey(ctx context.Context, req AccountIDRequest) (resp bitbucket_org_orus_io_go_xbus_api.Account, err error)
	SetAPIKey(ctx context.Context, req AccountUpdateRequest) (resp bitbucket_org_orus_io_go_xbus_api.Account, err error)
	ResetCert(ctx context.Context, req AccountIDRequest) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForAccount = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Account",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForAccount = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Account",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForAccount = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Account",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForAccount = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Account",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// AccountHandler provides a NATS subscription handler that can serve a
// subscription using a given AccountServer implementation.
type AccountHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  AccountServer
}

func NewAccountHandler(ctx context.Context, nc nrpc.NatsConn, s AccountServer) *AccountHandler {
	return &AccountHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewAccountConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s AccountServer) *AccountHandler {
	return &AccountHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *AccountHandler) Subject() string {
	return "xbus.control.account.>"
}

func (h *AccountHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.control", 0, "account", 0, msg.Subject)
	if err != nil {
		log.Printf("AccountHanlder: Account subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "list":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("ListHanlder: List subject parsing failed: %v", err)
			break
		}
		var req AccountListRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ListHandler: List request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForAccount.WithLabelValues(
				"List", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.List(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "update":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("UpdateHanlder: Update subject parsing failed: %v", err)
			break
		}
		var req AccountUpdateRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("UpdateHandler: Update request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForAccount.WithLabelValues(
				"Update", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Update(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "delete":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("DeleteHanlder: Delete subject parsing failed: %v", err)
			break
		}
		var req AccountIDRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("DeleteHandler: Delete request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForAccount.WithLabelValues(
				"Delete", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				var innerResp nrpc.Void
				err := h.server.Delete(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "renew_apikey":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("RenewAPIKeyHanlder: RenewAPIKey subject parsing failed: %v", err)
			break
		}
		var req AccountIDRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("RenewAPIKeyHandler: RenewAPIKey request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForAccount.WithLabelValues(
				"RenewAPIKey", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.RenewAPIKey(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "get_apikey":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("GetAPIKeyHanlder: GetAPIKey subject parsing failed: %v", err)
			break
		}
		var req AccountIDRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("GetAPIKeyHandler: GetAPIKey request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForAccount.WithLabelValues(
				"GetAPIKey", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.GetAPIKey(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "set_apikey":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("SetAPIKeyHanlder: SetAPIKey subject parsing failed: %v", err)
			break
		}
		var req AccountUpdateRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("SetAPIKeyHandler: SetAPIKey request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForAccount.WithLabelValues(
				"SetAPIKey", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.SetAPIKey(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "resetcert":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("ResetCertHanlder: ResetCert subject parsing failed: %v", err)
			break
		}
		var req AccountIDRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ResetCertHandler: ResetCert request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForAccount.WithLabelValues(
				"ResetCert", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				var innerResp nrpc.Void
				err := h.server.ResetCert(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("AccountHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForAccount.WithLabelValues(
			"Account", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForAccount.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForAccount.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForAccount.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForAccount.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("AccountHandler: Account handler failed to publish the response: %s", err)
			serverRequestsForAccount.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForAccount.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type AccountClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewAccountClient(nc nrpc.NatsConn) *AccountClient {
	return &AccountClient{
		nc:      nc,
		PkgSubject: "xbus.control",
		Subject: "account",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *AccountClient) List(req AccountListRequest) (resp AccountListReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "list"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForAccount.WithLabelValues(
			"List", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForAccount.WithLabelValues("List").Observe(elapsed)
	clientCallsForAccount.WithLabelValues(
		"List", c.Encoding, "success").Inc()

	return
}

func (c *AccountClient) Update(req AccountUpdateRequest) (resp bitbucket_org_orus_io_go_xbus_api.Account, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "update"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForAccount.WithLabelValues(
			"Update", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForAccount.WithLabelValues("Update").Observe(elapsed)
	clientCallsForAccount.WithLabelValues(
		"Update", c.Encoding, "success").Inc()

	return
}

func (c *AccountClient) Delete(req AccountIDRequest) (err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "delete"

	// call
	var resp github_com_nats_rpc_gogo_nrpc.Void
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForAccount.WithLabelValues(
			"Delete", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForAccount.WithLabelValues("Delete").Observe(elapsed)
	clientCallsForAccount.WithLabelValues(
		"Delete", c.Encoding, "success").Inc()

	return
}

func (c *AccountClient) RenewAPIKey(req AccountIDRequest) (resp bitbucket_org_orus_io_go_xbus_api.Account, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "renew_apikey"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForAccount.WithLabelValues(
			"RenewAPIKey", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForAccount.WithLabelValues("RenewAPIKey").Observe(elapsed)
	clientCallsForAccount.WithLabelValues(
		"RenewAPIKey", c.Encoding, "success").Inc()

	return
}

func (c *AccountClient) GetAPIKey(req AccountIDRequest) (resp bitbucket_org_orus_io_go_xbus_api.Account, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "get_apikey"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForAccount.WithLabelValues(
			"GetAPIKey", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForAccount.WithLabelValues("GetAPIKey").Observe(elapsed)
	clientCallsForAccount.WithLabelValues(
		"GetAPIKey", c.Encoding, "success").Inc()

	return
}

func (c *AccountClient) SetAPIKey(req AccountUpdateRequest) (resp bitbucket_org_orus_io_go_xbus_api.Account, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "set_apikey"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForAccount.WithLabelValues(
			"SetAPIKey", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForAccount.WithLabelValues("SetAPIKey").Observe(elapsed)
	clientCallsForAccount.WithLabelValues(
		"SetAPIKey", c.Encoding, "success").Inc()

	return
}

func (c *AccountClient) ResetCert(req AccountIDRequest) (err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "resetcert"

	// call
	var resp github_com_nats_rpc_gogo_nrpc.Void
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForAccount.WithLabelValues(
			"ResetCert", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForAccount.WithLabelValues("ResetCert").Observe(elapsed)
	clientCallsForAccount.WithLabelValues(
		"ResetCert", c.Encoding, "success").Inc()

	return
}

// ActorServer is the interface that providers of the service
// Actor should implement.
type ActorServer interface {
	Create(ctx context.Context, req ActorRequest) (resp ActorReply, err error)
	Update(ctx context.Context, req ActorRequest) (resp ActorReply, err error)
	Delete(ctx context.Context, req ActorRequest) (resp ActorReply, err error)
	Query(ctx context.Context, req ActorRequest) (resp ActorReply, err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForActor = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Actor",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForActor = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Actor",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForActor = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Actor",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForActor = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Actor",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// ActorHandler provides a NATS subscription handler that can serve a
// subscription using a given ActorServer implementation.
type ActorHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  ActorServer
}

func NewActorHandler(ctx context.Context, nc nrpc.NatsConn, s ActorServer) *ActorHandler {
	return &ActorHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewActorConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s ActorServer) *ActorHandler {
	return &ActorHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *ActorHandler) Subject() string {
	return "xbus.control.actor.>"
}

func (h *ActorHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.control", 0, "actor", 0, msg.Subject)
	if err != nil {
		log.Printf("ActorHanlder: Actor subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "create":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("CreateHanlder: Create subject parsing failed: %v", err)
			break
		}
		var req ActorRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("CreateHandler: Create request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForActor.WithLabelValues(
				"Create", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Create(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "update":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("UpdateHanlder: Update subject parsing failed: %v", err)
			break
		}
		var req ActorRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("UpdateHandler: Update request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForActor.WithLabelValues(
				"Update", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Update(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "delete":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("DeleteHanlder: Delete subject parsing failed: %v", err)
			break
		}
		var req ActorRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("DeleteHandler: Delete request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForActor.WithLabelValues(
				"Delete", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Delete(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "query":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("QueryHanlder: Query subject parsing failed: %v", err)
			break
		}
		var req ActorRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("QueryHandler: Query request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForActor.WithLabelValues(
				"Query", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Query(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("ActorHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForActor.WithLabelValues(
			"Actor", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForActor.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForActor.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForActor.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForActor.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("ActorHandler: Actor handler failed to publish the response: %s", err)
			serverRequestsForActor.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForActor.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type ActorClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewActorClient(nc nrpc.NatsConn) *ActorClient {
	return &ActorClient{
		nc:      nc,
		PkgSubject: "xbus.control",
		Subject: "actor",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *ActorClient) Create(req ActorRequest) (resp ActorReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "create"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForActor.WithLabelValues(
			"Create", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForActor.WithLabelValues("Create").Observe(elapsed)
	clientCallsForActor.WithLabelValues(
		"Create", c.Encoding, "success").Inc()

	return
}

func (c *ActorClient) Update(req ActorRequest) (resp ActorReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "update"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForActor.WithLabelValues(
			"Update", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForActor.WithLabelValues("Update").Observe(elapsed)
	clientCallsForActor.WithLabelValues(
		"Update", c.Encoding, "success").Inc()

	return
}

func (c *ActorClient) Delete(req ActorRequest) (resp ActorReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "delete"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForActor.WithLabelValues(
			"Delete", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForActor.WithLabelValues("Delete").Observe(elapsed)
	clientCallsForActor.WithLabelValues(
		"Delete", c.Encoding, "success").Inc()

	return
}

func (c *ActorClient) Query(req ActorRequest) (resp ActorReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "query"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForActor.WithLabelValues(
			"Query", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForActor.WithLabelValues("Query").Observe(elapsed)
	clientCallsForActor.WithLabelValues(
		"Query", c.Encoding, "success").Inc()

	return
}

// EnvelopeServer is the interface that providers of the service
// Envelope should implement.
type EnvelopeServer interface {
	Purge(ctx context.Context, pushRep func(TaskProgress)) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForEnvelope = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Envelope",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForEnvelope = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Envelope",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForEnvelope = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Envelope",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForEnvelope = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Envelope",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// EnvelopeHandler provides a NATS subscription handler that can serve a
// subscription using a given EnvelopeServer implementation.
type EnvelopeHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  EnvelopeServer
}

func NewEnvelopeHandler(ctx context.Context, nc nrpc.NatsConn, s EnvelopeServer) *EnvelopeHandler {
	return &EnvelopeHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewEnvelopeConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s EnvelopeServer) *EnvelopeHandler {
	return &EnvelopeHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *EnvelopeHandler) Subject() string {
	return "xbus.control.envelope.>"
}

func (h *EnvelopeHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.control", 0, "envelope", 0, msg.Subject)
	if err != nil {
		log.Printf("EnvelopeHanlder: Envelope subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "purge":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("PurgeHanlder: Purge subject parsing failed: %v", err)
			break
		}
		var req github_com_nats_rpc_gogo_nrpc.Void
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("PurgeHandler: Purge request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForEnvelope.WithLabelValues(
				"Purge", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.EnableStreamedReply()
			request.Handler = func(ctx context.Context)(proto.Message, error){
				err := h.server.Purge(ctx, func(rep TaskProgress){
					request.SendStreamReply(&rep)
				})
				return nil, err
			}
		}
	default:
		log.Printf("EnvelopeHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForEnvelope.WithLabelValues(
			"Envelope", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForEnvelope.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForEnvelope.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForEnvelope.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForEnvelope.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("EnvelopeHandler: Envelope handler failed to publish the response: %s", err)
			serverRequestsForEnvelope.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForEnvelope.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type EnvelopeClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewEnvelopeClient(nc nrpc.NatsConn) *EnvelopeClient {
	return &EnvelopeClient{
		nc:      nc,
		PkgSubject: "xbus.control",
		Subject: "envelope",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *EnvelopeClient) Purge(
	ctx context.Context,
	cb func (context.Context, TaskProgress),
) error {
	start := time.Now()
	subject := c.PkgSubject + "." + c.Subject + "." + "purge"

	sub, err := nrpc.StreamCall(ctx, c.nc, subject, &nrpc.Void{}, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForEnvelope.WithLabelValues(
			"Purge", c.Encoding, "error").Inc()
		return err
	}

	var res TaskProgress
	for {
		err = sub.Next(&res)
		if err != nil {
			break
		}
		cb(ctx, res)
	}
	if err == nrpc.ErrEOS {
		err = nil
	}
	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForEnvelope.WithLabelValues("Purge").Observe(elapsed)
	clientCallsForEnvelope.WithLabelValues(
		"Purge", c.Encoding, "success").Inc()
	return err
}

// PipelineServer is the interface that providers of the service
// Pipeline should implement.
type PipelineServer interface {
	Query(ctx context.Context, req PipelineQueryRequest) (resp PipelineQueryReply, err error)
	Save(ctx context.Context, req PipelineSaveRequest) (resp PipelineSaveReply, err error)
	SetStatus(ctx context.Context, req bitbucket_org_orus_io_go_xbus_api.PipelineInfo) (resp PipelineSetStatusReply, err error)
	GetGraph(ctx context.Context, req bitbucket_org_orus_io_go_xbus_api.PipelineInfo) (resp bitbucket_org_orus_io_go_xbus_api.StringReply, err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForPipeline = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Pipeline",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForPipeline = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Pipeline",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForPipeline = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Pipeline",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForPipeline = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Pipeline",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// PipelineHandler provides a NATS subscription handler that can serve a
// subscription using a given PipelineServer implementation.
type PipelineHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  PipelineServer
}

func NewPipelineHandler(ctx context.Context, nc nrpc.NatsConn, s PipelineServer) *PipelineHandler {
	return &PipelineHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewPipelineConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s PipelineServer) *PipelineHandler {
	return &PipelineHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *PipelineHandler) Subject() string {
	return "xbus.control.pipeline.>"
}

func (h *PipelineHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.control", 0, "pipeline", 0, msg.Subject)
	if err != nil {
		log.Printf("PipelineHanlder: Pipeline subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "query":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("QueryHanlder: Query subject parsing failed: %v", err)
			break
		}
		var req PipelineQueryRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("QueryHandler: Query request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForPipeline.WithLabelValues(
				"Query", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Query(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "save":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("SaveHanlder: Save subject parsing failed: %v", err)
			break
		}
		var req PipelineSaveRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("SaveHandler: Save request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForPipeline.WithLabelValues(
				"Save", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Save(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "setstatus":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("SetStatusHanlder: SetStatus subject parsing failed: %v", err)
			break
		}
		var req bitbucket_org_orus_io_go_xbus_api.PipelineInfo
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("SetStatusHandler: SetStatus request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForPipeline.WithLabelValues(
				"SetStatus", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.SetStatus(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "getgraph":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("GetGraphHanlder: GetGraph subject parsing failed: %v", err)
			break
		}
		var req bitbucket_org_orus_io_go_xbus_api.PipelineInfo
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("GetGraphHandler: GetGraph request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForPipeline.WithLabelValues(
				"GetGraph", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.GetGraph(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("PipelineHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForPipeline.WithLabelValues(
			"Pipeline", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForPipeline.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForPipeline.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForPipeline.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForPipeline.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("PipelineHandler: Pipeline handler failed to publish the response: %s", err)
			serverRequestsForPipeline.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForPipeline.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type PipelineClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewPipelineClient(nc nrpc.NatsConn) *PipelineClient {
	return &PipelineClient{
		nc:      nc,
		PkgSubject: "xbus.control",
		Subject: "pipeline",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *PipelineClient) Query(req PipelineQueryRequest) (resp PipelineQueryReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "query"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForPipeline.WithLabelValues(
			"Query", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForPipeline.WithLabelValues("Query").Observe(elapsed)
	clientCallsForPipeline.WithLabelValues(
		"Query", c.Encoding, "success").Inc()

	return
}

func (c *PipelineClient) Save(req PipelineSaveRequest) (resp PipelineSaveReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "save"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForPipeline.WithLabelValues(
			"Save", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForPipeline.WithLabelValues("Save").Observe(elapsed)
	clientCallsForPipeline.WithLabelValues(
		"Save", c.Encoding, "success").Inc()

	return
}

func (c *PipelineClient) SetStatus(req bitbucket_org_orus_io_go_xbus_api.PipelineInfo) (resp PipelineSetStatusReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "setstatus"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForPipeline.WithLabelValues(
			"SetStatus", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForPipeline.WithLabelValues("SetStatus").Observe(elapsed)
	clientCallsForPipeline.WithLabelValues(
		"SetStatus", c.Encoding, "success").Inc()

	return
}

func (c *PipelineClient) GetGraph(req bitbucket_org_orus_io_go_xbus_api.PipelineInfo) (resp bitbucket_org_orus_io_go_xbus_api.StringReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "getgraph"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForPipeline.WithLabelValues(
			"GetGraph", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForPipeline.WithLabelValues("GetGraph").Observe(elapsed)
	clientCallsForPipeline.WithLabelValues(
		"GetGraph", c.Encoding, "success").Inc()

	return
}

// LogsServer is the interface that providers of the service
// Logs should implement.
type LogsServer interface {
	Purge(ctx context.Context, req LogsPurgeRequest, pushRep func(LogsPurgeReply)) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForLogs = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Logs",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForLogs = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Logs",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForLogs = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Logs",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForLogs = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Logs",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// LogsHandler provides a NATS subscription handler that can serve a
// subscription using a given LogsServer implementation.
type LogsHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  LogsServer
}

func NewLogsHandler(ctx context.Context, nc nrpc.NatsConn, s LogsServer) *LogsHandler {
	return &LogsHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewLogsConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s LogsServer) *LogsHandler {
	return &LogsHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *LogsHandler) Subject() string {
	return "xbus.control.logs.>"
}

func (h *LogsHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.control", 0, "logs", 0, msg.Subject)
	if err != nil {
		log.Printf("LogsHanlder: Logs subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "purge":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("PurgeHanlder: Purge subject parsing failed: %v", err)
			break
		}
		var req LogsPurgeRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("PurgeHandler: Purge request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForLogs.WithLabelValues(
				"Purge", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.EnableStreamedReply()
			request.Handler = func(ctx context.Context)(proto.Message, error){
				err := h.server.Purge(ctx, req, func(rep LogsPurgeReply){
					request.SendStreamReply(&rep)
				})
				return nil, err
			}
		}
	default:
		log.Printf("LogsHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForLogs.WithLabelValues(
			"Logs", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForLogs.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForLogs.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForLogs.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForLogs.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("LogsHandler: Logs handler failed to publish the response: %s", err)
			serverRequestsForLogs.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForLogs.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type LogsClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewLogsClient(nc nrpc.NatsConn) *LogsClient {
	return &LogsClient{
		nc:      nc,
		PkgSubject: "xbus.control",
		Subject: "logs",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *LogsClient) Purge(
	ctx context.Context,
	req LogsPurgeRequest,
	cb func (context.Context, LogsPurgeReply),
) error {
	start := time.Now()
	subject := c.PkgSubject + "." + c.Subject + "." + "purge"

	sub, err := nrpc.StreamCall(ctx, c.nc, subject, &req, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForLogs.WithLabelValues(
			"Purge", c.Encoding, "error").Inc()
		return err
	}

	var res LogsPurgeReply
	for {
		err = sub.Next(&res)
		if err != nil {
			break
		}
		cb(ctx, res)
	}
	if err == nrpc.ErrEOS {
		err = nil
	}
	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForLogs.WithLabelValues("Purge").Observe(elapsed)
	clientCallsForLogs.WithLabelValues(
		"Purge", c.Encoding, "success").Inc()
	return err
}

// PMProcessServer is the interface that providers of the service
// PMProcess should implement.
type PMProcessServer interface {
	Query(ctx context.Context, req PMProcessQueryRequest) (resp PMProcessQueryReply, err error)
	SetStatus(ctx context.Context, req PMProcessSetStatusRequest) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForPMProcess = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "PMProcess",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForPMProcess = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "PMProcess",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForPMProcess = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "PMProcess",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForPMProcess = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "PMProcess",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// PMProcessHandler provides a NATS subscription handler that can serve a
// subscription using a given PMProcessServer implementation.
type PMProcessHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  PMProcessServer
}

func NewPMProcessHandler(ctx context.Context, nc nrpc.NatsConn, s PMProcessServer) *PMProcessHandler {
	return &PMProcessHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewPMProcessConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s PMProcessServer) *PMProcessHandler {
	return &PMProcessHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *PMProcessHandler) Subject() string {
	return "xbus.control.pm.>"
}

func (h *PMProcessHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.control", 0, "pm", 0, msg.Subject)
	if err != nil {
		log.Printf("PMProcessHanlder: PMProcess subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "query":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("QueryHanlder: Query subject parsing failed: %v", err)
			break
		}
		var req PMProcessQueryRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("QueryHandler: Query request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForPMProcess.WithLabelValues(
				"Query", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Query(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "setstatus":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("SetStatusHanlder: SetStatus subject parsing failed: %v", err)
			break
		}
		var req PMProcessSetStatusRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("SetStatusHandler: SetStatus request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForPMProcess.WithLabelValues(
				"SetStatus", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				var innerResp nrpc.Void
				err := h.server.SetStatus(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	default:
		log.Printf("PMProcessHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForPMProcess.WithLabelValues(
			"PMProcess", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForPMProcess.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForPMProcess.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForPMProcess.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForPMProcess.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("PMProcessHandler: PMProcess handler failed to publish the response: %s", err)
			serverRequestsForPMProcess.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForPMProcess.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type PMProcessClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewPMProcessClient(nc nrpc.NatsConn) *PMProcessClient {
	return &PMProcessClient{
		nc:      nc,
		PkgSubject: "xbus.control",
		Subject: "pm",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *PMProcessClient) Query(req PMProcessQueryRequest) (resp PMProcessQueryReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "query"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForPMProcess.WithLabelValues(
			"Query", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForPMProcess.WithLabelValues("Query").Observe(elapsed)
	clientCallsForPMProcess.WithLabelValues(
		"Query", c.Encoding, "success").Inc()

	return
}

func (c *PMProcessClient) SetStatus(req PMProcessSetStatusRequest) (err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "setstatus"

	// call
	var resp github_com_nats_rpc_gogo_nrpc.Void
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForPMProcess.WithLabelValues(
			"SetStatus", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForPMProcess.WithLabelValues("SetStatus").Observe(elapsed)
	clientCallsForPMProcess.WithLabelValues(
		"SetStatus", c.Encoding, "success").Inc()

	return
}

// ProcessServer is the interface that providers of the service
// Process should implement.
type ProcessServer interface {
	Query(ctx context.Context, req ProcessQueryRequest, pushRep func(ProcessQueryReply)) (err error)
	Control(ctx context.Context, req ProcessControlRequest) (err error)
	Export(ctx context.Context, req ProcessExportRequest) (resp ProcessExportReply, err error)
	ExportStream(ctx context.Context, req ProcessExportRequest, pushRep func(OctetStreamList)) (err error)
	Purge(ctx context.Context, req ProcessPurgeRequest, pushRep func(TaskProgress)) (err error)
}

var (
	// The request completion time, measured at client-side.
	clientRCTForProcess = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_client_request_completion_time_seconds",
			Help:       "The request completion time for calls, measured client-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Process",
			},
		},
		[]string{"method"})

	// The handler execution time, measured at server-side.
	serverHETForProcess = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "nrpc_server_handler_execution_time_seconds",
			Help:       "The handler execution time for calls, measured server-side.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "Process",
			},
		},
		[]string{"method"})

	// The counts of calls made by the client, classified by result type.
	clientCallsForProcess = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_client_calls_count",
			Help: "The count of calls made by the client.",
			ConstLabels: map[string]string{
				"service": "Process",
			},
		},
		[]string{"method", "encoding", "result_type"})

	// The counts of requests handled by the server, classified by result type.
	serverRequestsForProcess = prometheus.NewCounterVec(
		prometheus.CounterOpts{
			Name: "nrpc_server_requests_count",
			Help: "The count of requests handled by the server.",
			ConstLabels: map[string]string{
				"service": "Process",
			},
		},
		[]string{"method", "encoding", "result_type"})
)

// ProcessHandler provides a NATS subscription handler that can serve a
// subscription using a given ProcessServer implementation.
type ProcessHandler struct {
	ctx     context.Context
	workers *nrpc.WorkerPool
	nc      nrpc.NatsConn
	server  ProcessServer
}

func NewProcessHandler(ctx context.Context, nc nrpc.NatsConn, s ProcessServer) *ProcessHandler {
	return &ProcessHandler{
		ctx:    ctx,
		nc:     nc,
		server: s,
	}
}

func NewProcessConcurrentHandler(workers *nrpc.WorkerPool, nc nrpc.NatsConn, s ProcessServer) *ProcessHandler {
	return &ProcessHandler{
		workers: workers,
		nc:      nc,
		server:  s,
	}
}

func (h *ProcessHandler) Subject() string {
	return "xbus.control.process.>"
}

func (h *ProcessHandler) Handler(msg *nats.Msg) {
	var ctx context.Context
	if h.workers != nil {
		ctx = h.workers.Context
	} else {
		ctx = h.ctx
	}
	request := nrpc.NewRequest(ctx, h.nc, msg.Subject, msg.Reply)
	// extract method name & encoding from subject
	_, _, name, tail, err := nrpc.ParseSubject(
		"xbus.control", 0, "process", 0, msg.Subject)
	if err != nil {
		log.Printf("ProcessHanlder: Process subject parsing failed: %v", err)
		return
	}

	request.MethodName = name
	request.SubjectTail = tail

	// call handler and form response
	var immediateError *nrpc.Error
	switch name {
	case "query":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("QueryHanlder: Query subject parsing failed: %v", err)
			break
		}
		var req ProcessQueryRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("QueryHandler: Query request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForProcess.WithLabelValues(
				"Query", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.EnableStreamedReply()
			request.Handler = func(ctx context.Context)(proto.Message, error){
				err := h.server.Query(ctx, req, func(rep ProcessQueryReply){
					request.SendStreamReply(&rep)
				})
				return nil, err
			}
		}
	case "control":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("ControlHanlder: Control subject parsing failed: %v", err)
			break
		}
		var req ProcessControlRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ControlHandler: Control request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForProcess.WithLabelValues(
				"Control", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				var innerResp nrpc.Void
				err := h.server.Control(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "export":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("ExportHanlder: Export subject parsing failed: %v", err)
			break
		}
		var req ProcessExportRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ExportHandler: Export request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForProcess.WithLabelValues(
				"Export", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.Handler = func(ctx context.Context)(proto.Message, error){
				innerResp, err := h.server.Export(ctx, req)
				if err != nil {
					return nil, err
				}
				return &innerResp, err
			}
		}
	case "exportstream":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("ExportStreamHanlder: ExportStream subject parsing failed: %v", err)
			break
		}
		var req ProcessExportRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("ExportStreamHandler: ExportStream request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForProcess.WithLabelValues(
				"ExportStream", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.EnableStreamedReply()
			request.Handler = func(ctx context.Context)(proto.Message, error){
				err := h.server.ExportStream(ctx, req, func(rep OctetStreamList){
					request.SendStreamReply(&rep)
				})
				return nil, err
			}
		}
	case "purge":
		_, request.Encoding, err = nrpc.ParseSubjectTail(0, request.SubjectTail)
		if err != nil {
			log.Printf("PurgeHanlder: Purge subject parsing failed: %v", err)
			break
		}
		var req ProcessPurgeRequest
		if err := nrpc.Unmarshal(request.Encoding, msg.Data, &req); err != nil {
			log.Printf("PurgeHandler: Purge request unmarshal failed: %v", err)
			immediateError = &nrpc.Error{
				Type: nrpc.Error_CLIENT,
				Message: "bad request received: " + err.Error(),
			}
			serverRequestsForProcess.WithLabelValues(
				"Purge", request.Encoding, "unmarshal_fail").Inc()
		} else {
			request.EnableStreamedReply()
			request.Handler = func(ctx context.Context)(proto.Message, error){
				err := h.server.Purge(ctx, req, func(rep TaskProgress){
					request.SendStreamReply(&rep)
				})
				return nil, err
			}
		}
	default:
		log.Printf("ProcessHandler: unknown name %q", name)
		immediateError = &nrpc.Error{
			Type: nrpc.Error_CLIENT,
			Message: "unknown name: " + name,
		}
		serverRequestsForProcess.WithLabelValues(
			"Process", request.Encoding, "name_fail").Inc()
	}
	request.AfterReply = func(request *nrpc.Request, success, replySuccess bool) {
		if !replySuccess {
			serverRequestsForProcess.WithLabelValues(
				request.MethodName, request.Encoding, "sendreply_fail").Inc()
		}
		if success {
			serverRequestsForProcess.WithLabelValues(
				request.MethodName, request.Encoding, "success").Inc()
		} else {
			serverRequestsForProcess.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		// report metric to Prometheus
		serverHETForProcess.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	}
	if immediateError == nil {
		if h.workers != nil {
			// Try queuing the request
			if err := h.workers.QueueRequest(request); err != nil {
				log.Printf("nrpc: Error queuing the request: %s", err)
			}
		} else {
			// Run the handler synchronously
			request.RunAndReply()
		}
	}

	if immediateError != nil {
		if err := request.SendReply(nil, immediateError); err != nil {
			log.Printf("ProcessHandler: Process handler failed to publish the response: %s", err)
			serverRequestsForProcess.WithLabelValues(
				request.MethodName, request.Encoding, "handler_fail").Inc()
		}
		serverHETForProcess.WithLabelValues(request.MethodName).Observe(
			request.Elapsed().Seconds())
	} else {
	}
}

type ProcessClient struct {
	nc      nrpc.NatsConn
	PkgSubject string
	Subject string
	Encoding string
	Timeout time.Duration
}

func NewProcessClient(nc nrpc.NatsConn) *ProcessClient {
	return &ProcessClient{
		nc:      nc,
		PkgSubject: "xbus.control",
		Subject: "process",
		Encoding: "protobuf",
		Timeout: 5 * time.Second,
	}
}

func (c *ProcessClient) Query(
	ctx context.Context,
	req ProcessQueryRequest,
	cb func (context.Context, ProcessQueryReply),
) error {
	start := time.Now()
	subject := c.PkgSubject + "." + c.Subject + "." + "query"

	sub, err := nrpc.StreamCall(ctx, c.nc, subject, &req, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForProcess.WithLabelValues(
			"Query", c.Encoding, "error").Inc()
		return err
	}

	var res ProcessQueryReply
	for {
		err = sub.Next(&res)
		if err != nil {
			break
		}
		cb(ctx, res)
	}
	if err == nrpc.ErrEOS {
		err = nil
	}
	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForProcess.WithLabelValues("Query").Observe(elapsed)
	clientCallsForProcess.WithLabelValues(
		"Query", c.Encoding, "success").Inc()
	return err
}

func (c *ProcessClient) Control(req ProcessControlRequest) (err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "control"

	// call
	var resp github_com_nats_rpc_gogo_nrpc.Void
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForProcess.WithLabelValues(
			"Control", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForProcess.WithLabelValues("Control").Observe(elapsed)
	clientCallsForProcess.WithLabelValues(
		"Control", c.Encoding, "success").Inc()

	return
}

func (c *ProcessClient) Export(req ProcessExportRequest) (resp ProcessExportReply, err error) {
	start := time.Now()

	subject := c.PkgSubject + "." + c.Subject + "." + "export"

	// call
	err = nrpc.Call(&req, &resp, c.nc, subject, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForProcess.WithLabelValues(
			"Export", c.Encoding, "call_fail").Inc()
		return // already logged
	}

	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForProcess.WithLabelValues("Export").Observe(elapsed)
	clientCallsForProcess.WithLabelValues(
		"Export", c.Encoding, "success").Inc()

	return
}

func (c *ProcessClient) ExportStream(
	ctx context.Context,
	req ProcessExportRequest,
	cb func (context.Context, OctetStreamList),
) error {
	start := time.Now()
	subject := c.PkgSubject + "." + c.Subject + "." + "exportstream"

	sub, err := nrpc.StreamCall(ctx, c.nc, subject, &req, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForProcess.WithLabelValues(
			"ExportStream", c.Encoding, "error").Inc()
		return err
	}

	var res OctetStreamList
	for {
		err = sub.Next(&res)
		if err != nil {
			break
		}
		cb(ctx, res)
	}
	if err == nrpc.ErrEOS {
		err = nil
	}
	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForProcess.WithLabelValues("ExportStream").Observe(elapsed)
	clientCallsForProcess.WithLabelValues(
		"ExportStream", c.Encoding, "success").Inc()
	return err
}

func (c *ProcessClient) Purge(
	ctx context.Context,
	req ProcessPurgeRequest,
	cb func (context.Context, TaskProgress),
) error {
	start := time.Now()
	subject := c.PkgSubject + "." + c.Subject + "." + "purge"

	sub, err := nrpc.StreamCall(ctx, c.nc, subject, &req, c.Encoding, c.Timeout)
	if err != nil {
		clientCallsForProcess.WithLabelValues(
			"Purge", c.Encoding, "error").Inc()
		return err
	}

	var res TaskProgress
	for {
		err = sub.Next(&res)
		if err != nil {
			break
		}
		cb(ctx, res)
	}
	if err == nrpc.ErrEOS {
		err = nil
	}
	// report total time taken to Prometheus
	elapsed := time.Since(start).Seconds()
	clientRCTForProcess.WithLabelValues("Purge").Observe(elapsed)
	clientCallsForProcess.WithLabelValues(
		"Purge", c.Encoding, "success").Inc()
	return err
}

type Client struct {
	nc      nrpc.NatsConn
	defaultEncoding string
	defaultTimeout time.Duration
	pkgSubject string
	Account *AccountClient
	Actor *ActorClient
	Envelope *EnvelopeClient
	Pipeline *PipelineClient
	Logs *LogsClient
	PMProcess *PMProcessClient
	Process *ProcessClient
}

func NewClient(nc nrpc.NatsConn) *Client {
	c := Client{
		nc: nc,
		defaultEncoding: "protobuf",
		defaultTimeout: 5*time.Second,
		pkgSubject: "xbus.control",
	}
	c.Account = NewAccountClient(nc)
	c.Actor = NewActorClient(nc)
	c.Envelope = NewEnvelopeClient(nc)
	c.Pipeline = NewPipelineClient(nc)
	c.Logs = NewLogsClient(nc)
	c.PMProcess = NewPMProcessClient(nc)
	c.Process = NewProcessClient(nc)
	return &c
}

func (c *Client) SetEncoding(encoding string) {
	c.defaultEncoding = encoding
	if c.Account != nil {
		c.Account.Encoding = encoding
	}
	if c.Actor != nil {
		c.Actor.Encoding = encoding
	}
	if c.Envelope != nil {
		c.Envelope.Encoding = encoding
	}
	if c.Pipeline != nil {
		c.Pipeline.Encoding = encoding
	}
	if c.Logs != nil {
		c.Logs.Encoding = encoding
	}
	if c.PMProcess != nil {
		c.PMProcess.Encoding = encoding
	}
	if c.Process != nil {
		c.Process.Encoding = encoding
	}
}

func (c *Client) SetTimeout(t time.Duration) {
	c.defaultTimeout = t
	if c.Account != nil {
		c.Account.Timeout = t
	}
	if c.Actor != nil {
		c.Actor.Timeout = t
	}
	if c.Envelope != nil {
		c.Envelope.Timeout = t
	}
	if c.Pipeline != nil {
		c.Pipeline.Timeout = t
	}
	if c.Logs != nil {
		c.Logs.Timeout = t
	}
	if c.PMProcess != nil {
		c.PMProcess.Timeout = t
	}
	if c.Process != nil {
		c.Process.Timeout = t
	}
}

func init() {
	// register metrics for service Account
	prometheus.MustRegister(clientRCTForAccount)
	prometheus.MustRegister(serverHETForAccount)
	prometheus.MustRegister(clientCallsForAccount)
	prometheus.MustRegister(serverRequestsForAccount)
	// register metrics for service Actor
	prometheus.MustRegister(clientRCTForActor)
	prometheus.MustRegister(serverHETForActor)
	prometheus.MustRegister(clientCallsForActor)
	prometheus.MustRegister(serverRequestsForActor)
	// register metrics for service Envelope
	prometheus.MustRegister(clientRCTForEnvelope)
	prometheus.MustRegister(serverHETForEnvelope)
	prometheus.MustRegister(clientCallsForEnvelope)
	prometheus.MustRegister(serverRequestsForEnvelope)
	// register metrics for service Pipeline
	prometheus.MustRegister(clientRCTForPipeline)
	prometheus.MustRegister(serverHETForPipeline)
	prometheus.MustRegister(clientCallsForPipeline)
	prometheus.MustRegister(serverRequestsForPipeline)
	// register metrics for service Logs
	prometheus.MustRegister(clientRCTForLogs)
	prometheus.MustRegister(serverHETForLogs)
	prometheus.MustRegister(clientCallsForLogs)
	prometheus.MustRegister(serverRequestsForLogs)
	// register metrics for service PMProcess
	prometheus.MustRegister(clientRCTForPMProcess)
	prometheus.MustRegister(serverHETForPMProcess)
	prometheus.MustRegister(clientCallsForPMProcess)
	prometheus.MustRegister(serverRequestsForPMProcess)
	// register metrics for service Process
	prometheus.MustRegister(clientRCTForProcess)
	prometheus.MustRegister(serverHETForProcess)
	prometheus.MustRegister(clientCallsForProcess)
	prometheus.MustRegister(serverRequestsForProcess)
}